import torch


class Model(torch.nn.Module):

    def __init__(self, config, name):
        super(Model, self).__init__()
        self._config = config
        self._name = name

    def forward(self):
        raise NotImplementedError

    def __str__(self):
        return f'{self.__class__.__name__}(name={self._name})'

    @property
    def config(self):
        return self._config

    @property
    def name(self):
        return self._name



from types import SimpleNamespace
from typing import Mapping
from typing import Sequence

import torch
import torch.nn as nn
import torch.optim as optim
import wandb
from tqdm.autonotebook import tqdm

from nam.config import Config
from nam.models.saver import Checkpointer
from nam.trainer.losses import penalized_loss
from nam.trainer.metrics import accuracy
from nam.trainer.metrics import mae
from nam.utils.loggers import TensorBoardLogger


class Trainer:

    def __init__(self, config: SimpleNamespace, model: Sequence[nn.Module], dataset: torch.utils.data.Dataset) -> None:
        self.config = Config(**vars(config))  #config
        self.model = model
        self.dataset = dataset
        self.optimizer = torch.optim.Adam(self.model.parameters(),
                                          lr=self.config.lr,
                                          weight_decay=self.config.decay_rate)

        self.writer = TensorBoardLogger(config)
        self.checkpointer = Checkpointer(model=model, config=config)

        self.criterion = lambda inputs, targets, weights, fnns_out, model: penalized_loss(
            self.config, inputs, targets, weights, fnns_out, model)

        self.metrics = lambda logits, targets: mae(logits, targets) if config.regression else accuracy(logits, targets)
        self.metrics_name = "MAE" if config.regression else "Accuracy"

        if config.wandb:
            wandb.watch(models=self.model, log='all', log_freq=10)

        self.dataloader_train, self.dataloader_val, \
          self.dataloader_test = self.dataset.get_dataloaders()

    def train_step(self, model: nn.Module, optimizer: optim.Optimizer, batch: torch.Tensor) -> torch.Tensor:
        """Performs a single gradient-descent optimization step."""

        features, targets = batch

        # Resets optimizer's gradients.
        self.optimizer.zero_grad()

        # Forward pass from the model.
        predictions, fnn_out = self.model(features)

        loss = self.criterion(predictions, targets, None, fnn_out, self.model)
        metrics = self.metrics(predictions, targets)

        # Backward pass.
        loss.backward()

        # Performs a gradient descent step.
        self.optimizer.step()

        return loss, metrics

    def train_epoch(self, model: nn.Module, optimizer: optim.Optimizer,
                    dataloader: torch.utils.data.DataLoader) -> torch.Tensor:
        """Performs an epoch of gradient descent optimization on
        `dataloader`."""
        model.train()
        loss = 0.0
        metrics = 0.0
        with tqdm(dataloader, leave=False) as pbar:
            for batch in pbar:

                # Performs a gradient-descent step.
                step_loss, step_metrics = self.train_step(model, optimizer, batch)
                loss += step_loss
                metrics += step_metrics

                pbar.set_description(f"TL Step: {step_loss:.3f} | {self.metrics_name}: {step_metrics:.3f}")

        return loss / len(dataloader), metrics / len(dataloader)

    def evaluate_step(self, model: nn.Module, batch: Mapping[str, torch.Tensor]) -> torch.Tensor:
        """Evaluates `model` on a `batch`."""

        features, targets = batch

        # Forward pass from the model.
        predictions, fnn_out = self.model(features)

        # Calculates loss on mini-batch.
        loss = self.criterion(predictions, targets, None, fnn_out, self.model)
        metrics = self.metrics(predictions, targets)

        # self.writer.write({"val_loss_step": loss.detach().cpu().numpy().item()})

        return loss, metrics

    def evaluate_epoch(self, model: nn.Module, dataloader: torch.utils.data.DataLoader) -> torch.Tensor:
        """Performs an evaluation of the `model` on the `dataloader."""
        model.eval()
        loss = 0.0
        metrics = 0.0
        with tqdm(dataloader, leave=False) as pbar:
            for batch in pbar:
                # Accumulates loss in dataset.
                with torch.no_grad():
                    # step_loss = self.evaluate_step(model, batch, pbar)
                    # loss += self.evaluate_step(model, batch, pbar)
                    step_loss, step_metrics = self.evaluate_step(model, batch)
                    loss += step_loss
                    metrics += step_metrics

                    pbar.set_description((f"VL Step: {step_loss:.3f} | {self.metrics_name}: {step_metrics:.3f}"))

        return loss / len(dataloader), metrics / len(dataloader)

    def train(self):
        num_epochs = self.config.num_epochs

        with tqdm(range(num_epochs)) as pbar_epoch:
            for epoch in pbar_epoch:
                # Trains model on whole training dataset, and writes on `TensorBoard`.
                loss_train, metrics_train = self.train_epoch(self.model, self.optimizer, self.dataloader_train)
                self.writer.write({
                    "loss_train_epoch": loss_train.detach().cpu().numpy().item(),
                    f"{self.metrics_name}_train_epoch": metrics_train,
                })

                # Evaluates model on whole validation dataset, and writes on `TensorBoard`.
                loss_val, metrics_val = self.evaluate_epoch(self.model, self.dataloader_val)
                self.writer.write({
                    "loss_val_epoch": loss_val.detach().cpu().numpy().item(),
                    f"{self.metrics_name}_val_epoch": metrics_val,
                })

                # Checkpoints model weights.
                if epoch % self.config.save_model_frequency == 0:
                    self.checkpointer.save(epoch)

                # Updates progress bar description.
                pbar_epoch.set_description(f"""Epoch({epoch}):
            TL: {loss_train.detach().cpu().numpy().item():.3f} |
            VL: {loss_val.detach().cpu().numpy().item():.3f} |
            {self.metrics_name}: {metrics_train:.3f}""")

    def test(self):
        num_epochs = self.config.num_epochs

        with tqdm(range(num_epochs)) as pbar_epoch:
            for epoch in pbar_epoch:

                # Evaluates model on whole validation dataset, and writes on `TensorBoard`.
                loss_test, metrics_test = self.evaluate_epoch(self.model, self.dataloader_test)
                # tune.report(loss_test=loss_test.detach().cpu().numpy().item())
                self.writer.write({
                    "loss_test_epoch": loss_test.detach().cpu().numpy().item(),
                    f"{self.metrics_name}_test_epoch": metrics_test,
                })

                # Updates progress bar description.
                pbar_epoch.set_description("Test Loss: {:.2f} ".format(loss_test.detach().cpu().numpy().item()))



import torch
import torch.nn as nn
import torch.nn.functional as F

from nam.models.base import Model

from .activation import ExU
from .activation import LinReLU


class FeatureNN(Model):
    """Neural Network model for each individual feature."""

    def __init__(
        self,
        config,
        name,
        *,
        input_shape: int,
        num_units: int,
        feature_num: int = 0,
    ) -> None:
        """Initializes FeatureNN hyperparameters.

        Args:
          num_units: Number of hidden units in first hidden layer.
          dropout: Coefficient for dropout regularization.
          feature_num: Feature Index used for naming the hidden layers.
        """
        super(FeatureNN, self).__init__(config, name)
        self._input_shape = input_shape
        self._num_units = num_units
        self._feature_num = feature_num
        self.dropout = nn.Dropout(p=self.config.dropout)

        hidden_sizes = [self._num_units] + self.config.hidden_sizes

        layers = []

        ## First layer is ExU
        if self.config.activation == "exu":
            layers.append(ExU(in_features=input_shape, out_features=num_units))
        else:
            layers.append(LinReLU(in_features=input_shape, out_features=num_units))

        ## Hidden Layers
        for in_features, out_features in zip(hidden_sizes, hidden_sizes[1:]):
            layers.append(LinReLU(in_features, out_features))

        ## Last Linear Layer
        layers.append(nn.Linear(in_features=hidden_sizes[-1], out_features=1))

        self.model = nn.ModuleList(layers)
        # self.apply(init_weights)

    def forward(self, inputs) -> torch.Tensor:
        """Computes FeatureNN output with either evaluation or training
        mode."""
        outputs = inputs.unsqueeze(1)
        for layer in self.model:
            outputs = self.dropout(layer(outputs))
        return outputs



import argparse


def parse_args() -> argparse.Namespace:
    parser = argparse.ArgumentParser("Neural Additive Models")

    parser.add_argument('--num_epochs', default=1, type=int, help='The number of epochs to run training for.')

    parser.add_argument('--learning_rate', default=1e-2, type=float, help='Hyperparameter: learning rate.')

    parser.add_argument('--batch_size', default=1024, type=int, help='Hyperparameter: batch size.')

    parser.add_argument('--data_path', required=True, type=str, help='The path for the training data')

    parser.add_argument('--features_columns',
                        required=True,
                        nargs='+',
                        help='Name of the feature columns in the dataset')

    parser.add_argument('--targets_column', required=True, nargs='+', help='Name of the target column in the dataset')

    parser.add_argument('--weights_column', default="wgts", help='Name of the weights column in the dataset')

    parser.add_argument('--experiment_name', default="NAM_Experiment", type=str, help='The name for the experiment')

    parser.add_argument('--regression',
                        default=False,
                        type=bool,
                        help='Boolean flag indicating whether we '
                        'are solving a regression task or a classification task.')

    parser.add_argument('--logdir', default="output", type=str, help='Path to dir where to store summaries.')

    parser.add_argument('--wandb', default=False, type=bool, help='Using wandb for experiments tracking and logging')

    parser.add_argument('--hidden_sizes', default=[], type=int, nargs='+', help='Feature Neural Net hidden sizes')

    parser.add_argument('--activation',
                        default='exu',
                        type=str,
                        choices=['exu', 'relu'],
                        help='Activation function to used in the '
                        'hidden layer. Possible options: (1) relu, (2) exu')

    parser.add_argument('--dropout', default=0.5, type=float, help='Hyperparameter: Dropout rate')

    parser.add_argument('--feature_dropout',
                        default=0.5,
                        type=float,
                        help='Hyperparameter: Prob. with which '
                        'features are dropped')

    parser.add_argument('--decay_rate', default=0.995, type=float, help='Hyperparameter: Optimizer decay rate')

    parser.add_argument('--l2_regularization', default=0.5, type=float, help='Hyperparameter: l2 weight decay')

    parser.add_argument('--output_regularization', default=0.5, type=float, help='Hyperparameter: feature reg')

    parser.add_argument('--dataset_name', default='Teleco', type=str, help='Name of the dataset to load for training.')

    parser.add_argument('--seed', default=2021, type=int, help='seed for torch')

    parser.add_argument('--num_basis_functions',
                        default=1000,
                        type=int,
                        help='Number of basis functions '
                        'to use in a FeatureNN for a real-valued feature.')

    parser.add_argument('--units_multiplier',
                        default=2,
                        type=int,
                        help='Number of basis functions for a '
                        'categorical feature')

    parser.add_argument('--shuffle', default=True, type=bool, help='Shuffle the training data')

    parser.add_argument('--cross_val',
                        default=False,
                        type=bool,
                        help='Boolean flag indicating whether to '
                        'perform cross validation or not.')

    parser.add_argument('--num_folds', default=5, type=int, help='Number of N folds')

    parser.add_argument('--num_splits', default=3, type=int, help='Number of data splits to use')

    parser.add_argument('--fold_num', default=1, type=int, help='Index of the fold to be used')

    parser.add_argument('--num_models', default=1, type=int, help='the number of models to train.')

    parser.add_argument('--early_stopping_patience', default=50, type=int, help='Early stopping epochs')

    parser.add_argument('--use_dnn', default=False, type=bool, help='Deep NN baseline.')

    parser.add_argument('--save_top_k',
                        default=3,
                        type=int,
                        help='Indicates the maximum '
                        'number of recent checkpoint files to keep.')

    FLAGS = parser.parse_args()
    return FLAGS



from typing import Union

import numpy as np
import pandas as pd
from sklearn.compose import ColumnTransformer
from sklearn.pipeline import Pipeline
from sklearn.preprocessing import FunctionTransformer
from sklearn.preprocessing import MinMaxScaler
from sklearn.preprocessing import OneHotEncoder


def missing_values_table(df):
    mis_val = df.isnull().sum()
    mis_val_percent = 100 * df.isnull().sum() / len(df)
    mis_val_table = pd.concat([mis_val, mis_val_percent], axis=1)
    mis_val_table_ren_columns = mis_val_table.rename(columns={0: 'Missing Values', 1: '% of Total Values'})
    mis_val_table_ren_columns = mis_val_table_ren_columns[mis_val_table_ren_columns.iloc[:, 1] != 0].sort_values(
        '% of Total Values', ascending=False).round(3)
    print("Your selected dataframe has " + str(df.shape[1]) + " columns.\n"
          "There are " + str(mis_val_table_ren_columns.shape[0]) + " columns that have missing values.")
    return mis_val_table_ren_columns


def handle_nan(df: pd.DataFrame) -> pd.DataFrame:
    ## TODO: Add more methods
    return df.interpolate(method='linear', axis=0)


def read_dataset(file_path: Union[str, pd.DataFrame],
                 header='infer',
                 names=None,
                 delim_whitespace=False) -> pd.DataFrame:
    if isinstance(file_path, str):
        data = pd.read_csv(
            file_path,
            header=header,
            names=names,
            delim_whitespace=delim_whitespace,
        )
    else:
        data = file_path

    ## Handle NANs
    if data.isnull().values.any():
        print('Found `Nulls` values in the dataset')
        print(missing_values_table(data))
        data = handle_nan(data)

    return data


class CustomPipeline(Pipeline):
    """Custom sklearn Pipeline to transform data."""

    def apply_transformation(self, inputs):
        """Applies all transforms to the data, without applying last estimator.

        Args:
          x: Iterable data to predict on. Must fulfill input requirements of first
            step of the pipeline.

        Returns:
          xt: Transformed data.
        """
        xt = inputs
        for _, transform in self.steps[:-1]:
            xt = transform.fit_transform(xt)
        return xt


def transform_data(df: pd.DataFrame):
    column_names = df.columns
    new_column_names = []

    is_categorical = np.array([dt.kind == 'O' for dt in df.dtypes])
    categorical_cols = df.columns.values[is_categorical]
    numerical_cols = df.columns.values[~is_categorical]

    for index, is_cat in enumerate(is_categorical):
        col_name = column_names[index]
        if is_cat:
            new_column_names.append([f'{col_name}_{val}' for val in set(df[col_name])])
        else:
            new_column_names.append(col_name)

    cat_ohe_step = ('ohe', OneHotEncoder(sparse=False, handle_unknown='ignore'))

    cat_pipe = Pipeline([cat_ohe_step])
    num_pipe = Pipeline([('identity', FunctionTransformer(validate=True))])
    transformers = [('cat', cat_pipe, categorical_cols), ('num', num_pipe, numerical_cols)]
    column_transform = ColumnTransformer(transformers=transformers)

    pipe = CustomPipeline([('column_transform', column_transform), ('min_max', MinMaxScaler((-1, 1))), ('dummy', None)])
    df = pipe.apply_transformation(df)
    return df, new_column_names



from typing import Tuple
from typing import Union

import numpy as np
import pandas as pd
import torch
from torch.utils.data import DataLoader
from torch.utils.data import random_split

from .utils import read_dataset
from .utils import transform_data


class CSVDataset(torch.utils.data.Dataset):

    def __init__(self,
                 config,
                 data_path: Union[str, pd.DataFrame],
                 features_columns: list,
                 targets_column: str,
                 weights_column: str = None):
        """Custom dataset for csv files.

        Args:
            config ([type]): [description]
            data_path (str): [description]
            features_columns (list): [description]
            targets_column (str): [description]
            weights_column (str, optional): [description]. Defaults to None.
        """

        self._config = config
        self.data_path = data_path
        self.features_columns = features_columns
        self.targets_column = targets_column
        self.weights_column = weights_column

        self.data = read_dataset(data_path)

        self.raw_X = self.data[features_columns].copy()
        self.raw_y = self.data[targets_column].copy()

        self.X, self.y = self.raw_X.to_numpy(), self.raw_y.to_numpy()
        if weights_column is not None:
            self.wgts = self.data[weights_column].copy().to_numpy()
        else:
            self.wgts = np.ones_like(self.raw_y)

    def __len__(self):
        return len(self.data)

    def __getitem__(self, idx: int) -> Tuple[np.array, ...]:
        return self.X[idx], self.y[idx], self.wgts[idx]

    @property
    def config(self):
        return self._config


class NAMDataset(CSVDataset):

    def __init__(self,
                 config,
                 data_path: Union[str, pd.DataFrame],
                 features_columns: list,
                 targets_column: str,
                 weights_column: str = None) -> None:
        super().__init__(config=config,
                         data_path=data_path,
                         features_columns=features_columns,
                         targets_column=targets_column,
                         weights_column=weights_column)

        self.col_min_max = self.get_col_min_max()

        self.features, self.features_names = transform_data(self.raw_X)
        self.compute_features()

        if (not config.regression) and (not isinstance(self.raw_y, np.ndarray)):
            targets = pd.get_dummies(self.raw_y).values
            targets = np.array(np.argmax(targets, axis=-1))
        else:
            targets = self.y

        self.features = torch.from_numpy(self.features).float().to(config.device)
        self.targets = torch.from_numpy(targets).view(-1, 1).float().to(config.device)
        self.wgts = torch.from_numpy(self.wgts).to(config.device)

        self.setup_dataloaders()

    def __len__(self):
        return len(self.features)

    def __getitem__(self, idx: int) -> Tuple[np.array, ...]:
        return self.features[idx], self.targets[idx]  #, self.wgts[idx]

    def get_col_min_max(self):
        col_min_max = {}
        for col in self.raw_X:
            unique_vals = self.raw_X[col].unique()
            col_min_max[col] = (np.min(unique_vals), np.max(unique_vals))

        return col_min_max

    def compute_features(self):
        single_features = np.split(np.array(self.features), self.features.shape[1], axis=1)
        self.unique_features = [np.unique(f, axis=0) for f in single_features]

        self.single_features = {col: sorted(self.raw_X[col].to_numpy()) for col in self.raw_X}
        self.ufo = {col: sorted(self.raw_X[col].unique()) for col in self.raw_X}

    def setup_dataloaders(self, val_split: float = 0.1, test_split: float = 0.2) -> Tuple[DataLoader, ...]:
        test_size = int(test_split * len(self))
        val_size = int(val_split * (len(self) - test_size))
        train_size = len(self) - val_size - test_size

        train_subset, val_subset, test_subset = random_split(self, [train_size, val_size, test_size])

        self.train_dl = DataLoader(train_subset, batch_size=self.config.batch_size, shuffle=True)

        self.val_dl = DataLoader(val_subset, batch_size=self.config.batch_size, shuffle=False)

        self.test_dl = DataLoader(test_subset, batch_size=self.config.batch_size, shuffle=False)

    def train_dataloaders(self) -> Tuple[DataLoader, ...]:
        yield self.train_dl, self.val_dl

    def test_dataloaders(self) -> DataLoader:
        return self.test_dl



"""Base logger, borrowed from DeepMind"s Acme."""

from typing import Any
from typing import Mapping

from loguru import logger

LoggingData = Mapping[str, Any]


class Logger:
    """A logger has a `write` method."""

    def __init__(
        self,
        log_dir: str,
    ) -> None:
        self._logger = logger
        self.info = self._logger.info
        self.trace = self._logger.trace
        self.debug = self._logger.debug
        self.warning = self._logger.warning
        self.success = self._logger.success
        self.error = self._logger.error
        self.critical = self._logger.critical
        self._logger.add(f"{log_dir}/out.log")

    def write(self, data: LoggingData):
        """Writes `data` to destination (file, terminal, database, etc)."""
        raise NotImplementedError

    @property
    def logger(self):
        return self._logger



"""Utility classes for saving model checkpoints."""

import os

import torch
import torch.nn as nn

device = torch.device("cuda" if torch.cuda.is_available() else "cpu")


class Checkpointer:
    """A simple `PyTorch` model load/save wrapper."""

    def __init__(
        self,
        model: nn.Module,
        config: dict,
    ) -> None:
        """Constructs a simple load/save checkpointer."""
        self._model = model
        self._config = config

        self._ckpt_dir = os.path.join(config.logdir, "ckpts")
        os.makedirs(self._ckpt_dir, exist_ok=True)

    def save(
        self,
        epoch: int,
    ) -> str:
        """Saves the model to the `ckpt_dir/epoch/model.pt` file."""
        ckpt_path = os.path.join(self._ckpt_dir, "model-{}.pt".format(epoch))
        torch.save(self._model.state_dict(), ckpt_path)
        return ckpt_path

    def load(
        self,
        epoch: int,
    ) -> nn.Module:
        """Loads the model from the `ckpt_dir/epoch/model.pt` file."""
        ckpt_path = os.path.join(self._ckpt_dir, "model-{}.pt".format(epoch))
        self._model.load_state_dict(torch.load(ckpt_path, map_location=device))
        return self._model



import torch
import torch.nn as nn
import torch.nn.functional as F

from nam.types import Config


def bce_loss(logits: torch.Tensor, targets: torch.Tensor, weights: torch.tensor) -> torch.Tensor:
    """Cross entropy loss for binary classification.

    Args:
      logits: NAM model outputs
      targets: Binary class labels.

    Returns:
      Binary Cross-entropy loss between model predictions and the targets.
    """
    return F.binary_cross_entropy_with_logits(logits.view(-1), targets.view(-1))


def mse_loss(logits: torch.Tensor, targets: torch.Tensor, weights: torch.tensor) -> torch.Tensor:
    """Mean squared error loss for regression."""
    return F.mse_loss(logits.view(-1), targets.view(-1))


def penalized_loss(config: Config, logits: torch.Tensor, targets: torch.Tensor, weights: torch.tensor,
                   fnn_out: torch.Tensor, model: nn.Module) -> torch.Tensor:
    """Computes penalized loss with L2 regularization and output penalty.

    Args:
      config: Global config.
      model: Neural network model.
      inputs: Input values to be fed into the model for computing predictions.
      targets: Target values containing either real values or binary labels.

    Returns:
      The penalized loss.
    """

    def features_loss(per_feature_outputs: torch.Tensor) -> torch.Tensor:
        """Penalizes the L2 norm of the prediction of each feature net."""
        return (config.output_regularization * (per_feature_outputs**2).sum() / len(per_feature_outputs))

    def weight_decay(model: nn.Module) -> torch.Tensor:
        """Penalizes the L2 norm of weights in each feature net."""
        num_networks = 1 if config.use_dnn else len(model.feature_nns)
        l2_losses = torch.as_tensor([((x**2).sum() / 2) for x in model.parameters()])
        return torch.sum(l2_losses) / num_networks

    loss_func = mse_loss if config.regression else bce_loss
    loss = loss_func(logits, targets, weights)

    reg_loss = 0.0
    if config.output_regularization > 0:
        reg_loss += features_loss(fnn_out)

    if config.l2_regularization > 0:
        reg_loss += config.l2_regularization * weight_decay(model)

    return loss + reg_loss



import torch

from nam.config.base import Config


def defaults() -> Config:
    config = Config(
        device='cuda' if torch.cuda.is_available() else 'cpu',
        seed=2021,

        ## Data Path
        data_path="data/GALLUP.csv",
        experiment_name="NAM",
        regression=False,

        ## training
        num_epochs=1,
        lr=3e-4,
        batch_size=1024,

        ## logs
        logdir="output",
        wandb=True,

        ## Hidden size for layers
        hidden_sizes=[64, 32],

        ## Activation choice
        activation='exu',  ## Either `ExU` or `Relu`
        optimizer='adam',

        ## regularization_techniques
        dropout=0.5,
        feature_dropout=0.5,
        decay_rate=0.995,
        l2_regularization=0.5,
        output_regularization=0.5,

        ## Num units for FeatureNN
        num_basis_functions=1000,
        units_multiplier=2,
        shuffle=True,

        ## Folded
        cross_val=False,
        num_folds=5,
        num_splits=3,
        fold_num=1,

        ## Models
        num_models=1,

        ## for dataloaders
        num_workers=16,

        ## saver
        save_model_frequency=2,
        save_top_k=3,

        ## Early stopping
        use_dnn=False,
        early_stopping_patience=50,  ## For early stopping
    )

    return config



def mae(logits, targets):
    return ((logits.view(-1) - targets.view(-1)).abs().sum() / logits.numel()).item()


def accuracy(logits, targets):
    return (((targets.view(-1) > 0) == (logits.view(-1) > 0.5)).sum() / targets.numel()).item()



from typing import Sequence
from typing import Tuple

import matplotlib.patches as patches
import matplotlib.pyplot as plt
import numpy as np
import torch


def get_feature_contributions(
    model: torch.nn.Module,
    dataset: torch.utils.data.Dataset,
) -> Sequence[torch.Tensor]:

    feature_contributions = []
    unique_features = dataset.unique_features
    for i, feature in enumerate(unique_features):
        feature = torch.tensor(feature).float().to(model.config.device)
        feat_contribution = model.feature_nns[i](feature).cpu().detach().numpy().squeeze()
        feature_contributions.append(feat_contribution)

    return feature_contributions


def calc_mean_prediction(
    model: torch.nn.Module,
    dataset: torch.utils.data.Dataset,
) -> Tuple[np.ndarray, dict]:
    #@title Calculate the mean prediction

    feature_contributions = get_feature_contributions(model, dataset)
    avg_hist_data = {col: contributions for col, contributions in zip(dataset.features_names, feature_contributions)}
    all_indices, mean_pred = {}, {}

    for i, col in enumerate(dataset.features_names):
        feature_i = dataset.features[:, i].cpu()
        all_indices[col] = np.searchsorted(dataset.unique_features[i][:, 0], feature_i, 'left')

    for col in dataset.features_names:
        mean_pred[col] = np.mean([avg_hist_data[col]])  #[i] for i in all_indices[col]]) TODO: check the error here

    return mean_pred, avg_hist_data


def plot_mean_feature_importance(model: torch.nn.Module, dataset: torch.utils.data.Dataset, width=0.5):

    mean_pred, avg_hist_data = calc_mean_prediction(model, dataset)

    def compute_mean_feature_importance(mean_pred, avg_hist_data):
        mean_abs_score = {}
        for k in avg_hist_data:
            try:
                mean_abs_score[k] = np.mean(np.abs(avg_hist_data[k] - mean_pred[k]))
            except:
                continue
        x1, x2 = zip(*mean_abs_score.items())
        return x1, x2

    ## TODO: rename x1 and x2
    x1, x2 = compute_mean_feature_importance(mean_pred, avg_hist_data)

    cols = dataset.features_names
    fig = plt.figure(figsize=(5, 5))
    ind = np.arange(len(x1))
    x1_indices = np.argsort(x2)

    cols_here = [cols[i] for i in x1_indices]
    x2_here = [x2[i] for i in x1_indices]

    plt.bar(ind, x2_here, width, label='NAMs')
    plt.xticks(ind + width / 2, cols_here, rotation=90, fontsize='large')
    plt.ylabel('Mean Absolute Score', fontsize='x-large')
    plt.legend(loc='upper right', fontsize='large')
    plt.title(f'Overall Importance', fontsize='x-large')
    plt.show()

    return fig


def plot_nams(model: torch.nn.Module,
              dataset: torch.utils.data.Dataset,
              num_cols: int = 2,
              n_blocks: int = 20,
              color: list = [0.4, 0.5, 0.9],
              linewidth: float = 7.0,
              alpha: float = 1.0,
              feature_to_use: list = None):

    unique_features, single_features = dataset.ufo, dataset.single_features
    mean_pred, feat_data_contrib = calc_mean_prediction(model, dataset)

    num_rows = len(dataset.features[0]) // num_cols

    fig = plt.figure(num=None, figsize=(num_cols * 10, num_rows * 10), facecolor='w', edgecolor='k')
    fig.tight_layout(pad=7.0)

    feat_data_contrib_pairs = list(feat_data_contrib.items())
    feat_data_contrib_pairs.sort(key=lambda x: x[0])

    mean_pred_pairs = list(mean_pred.items())
    mean_pred_pairs.sort(key=lambda x: x[0])

    if feature_to_use:
        feat_data_contrib_pairs = [v for v in feat_data_contrib_pairs if v[0] in feature_to_use]

    min_y = np.min([np.min(a[1]) for a in feat_data_contrib_pairs])
    max_y = np.max([np.max(a[1]) for a in feat_data_contrib_pairs])

    min_max_dif = max_y - min_y
    min_y = min_y - 0.1 * min_max_dif
    max_y = max_y + 0.1 * min_max_dif

    total_mean_bias = 0

    def shade_by_density_blocks(color: list = [0.9, 0.5, 0.5]):
        single_feature_data = single_features[name]
        x_n_blocks = min(n_blocks, len(unique_feat_data))

        segments = (max_x - min_x) / x_n_blocks
        density = np.histogram(single_feature_data, bins=x_n_blocks)
        normed_density = density[0] / np.max(density[0])
        rect_params = []

        for p in range(x_n_blocks):
            start_x = min_x + segments * p
            end_x = min_x + segments * (p + 1)
            d = min(1.0, 0.01 + normed_density[p])
            rect_params.append((d, start_x, end_x))

        for param in rect_params:
            alpha, start_x, end_x = param
            rect = patches.Rectangle(
                (start_x, min_y - 1),
                end_x - start_x,
                max_y - min_y + 1,
                linewidth=0.01,
                edgecolor=color,
                facecolor=color,
                alpha=alpha,
            )
            ax.add_patch(rect)

    for i, (name, feat_contrib) in enumerate(feat_data_contrib_pairs):
        mean_pred = mean_pred_pairs[i][1]
        total_mean_bias += mean_pred

        unique_feat_data = unique_features[name]
        ax = plt.subplot(num_rows, num_cols, i + 1)

        ## TODO: CATEGORICAL_NAMES if..else
        plt.plot(unique_feat_data, feat_contrib - mean_pred, color=color, linewidth=linewidth, alpha=alpha)

        plt.xticks(fontsize='x-large')

        plt.ylim(min_y, max_y)
        plt.yticks(fontsize='x-large')

        min_x = np.min(unique_feat_data)  # - 0.5  ## for categorical
        max_x = np.max(unique_feat_data)  # + 0.5
        plt.xlim(min_x, max_x)

        shade_by_density_blocks()

        if i % num_cols == 0:
            plt.ylabel('Features Contribution', fontsize='x-large')

        plt.xlabel(name, fontsize='x-large')

    plt.show()

    return fig



from typing import Sequence
from typing import Tuple

import torch
import torch.nn as nn
import torch.nn.functional as F

from nam.models.base import Model
from nam.models.featurenn import FeatureNN


class NAM(Model):

    def __init__(
        self,
        config,
        name,
        *,
        num_inputs: int,
        num_units: int,
    ) -> None:
        super(NAM, self).__init__(config, name)

        self._num_inputs = num_inputs
        self.dropout = nn.Dropout(p=self.config.dropout)

        if isinstance(num_units, list):
            assert len(num_units) == num_inputs
            self._num_units = num_units
        elif isinstance(num_units, int):
            self._num_units = [num_units for _ in range(self._num_inputs)]

        ## Builds the FeatureNNs on the first call.
        self.feature_nns = nn.ModuleList([
            FeatureNN(config=config, name=f'FeatureNN_{i}', input_shape=1, num_units=self._num_units[i], feature_num=i)
            for i in range(num_inputs)
        ])

        self._bias = torch.nn.Parameter(data=torch.zeros(1))

    def calc_outputs(self, inputs: torch.Tensor) -> Sequence[torch.Tensor]:
        """Returns the output computed by each feature net."""
        return [self.feature_nns[i](inputs[:, i]) for i in range(self._num_inputs)]

    def forward(self, inputs: torch.Tensor) -> Tuple[torch.Tensor, torch.Tensor]:
        individual_outputs = self.calc_outputs(inputs)
        conc_out = torch.cat(individual_outputs, dim=-1)
        dropout_out = self.dropout(conc_out)

        out = torch.sum(dropout_out, dim=-1)
        return out + self._bias, dropout_out



from typing import Dict

import numpy as np
import pandas as pd
import sklearn
from sklearn.datasets import load_breast_cancer

from nam.config import defaults
from nam.data.base import NAMDataset
from nam.data.folded import FoldedDataset

cfg = defaults()


def load_breast_data(config=cfg) -> Dict:
    breast_cancer = load_breast_cancer()
    dataset = pd.DataFrame(data=breast_cancer.data, columns=breast_cancer.feature_names)
    dataset['target'] = breast_cancer.target

    config.regression = False

    if config.cross_val:
        return FoldedDataset(config,
                             data_path=dataset,
                             features_columns=dataset.columns[:-1],
                             targets_column=dataset.columns[-1])
    else:
        return NAMDataset(config,
                          data_path=dataset,
                          features_columns=dataset.columns[:-1],
                          targets_column=dataset.columns[-1])


def load_sklearn_housing_data(config=cfg) -> Dict:
    housing = sklearn.datasets.fetch_california_housing()

    dataset = pd.DataFrame(data=housing.data, columns=housing.feature_names)
    dataset['target'] = housing.target

    config.regression = True

    if config.cross_val:
        return FoldedDataset(config,
                             data_path=dataset,
                             features_columns=dataset.columns[:-1],
                             targets_column=dataset.columns[-1])
    else:
        return NAMDataset(config,
                          data_path=dataset,
                          features_columns=dataset.columns[:-1],
                          targets_column=dataset.columns[-1])


def load_housing_data(config=cfg,
                      data_path: str = 'data/housing.csv',
                      features_columns: list = [
                          'longitude', 'latitude', 'housing_median_age', 'total_rooms', 'total_bedrooms', 'population',
                          'households', 'median_income'
                      ],
                      targets_column: str = 'median_house_value') -> Dict:

    config.regression = True
    feature_names = ['MedInc', 'HouseAge', 'AveRooms', 'AveBedrms', 'population', 'AveOccup', 'Latitude', 'Longitude']

    data = pd.read_csv(data_path)

    data['MedInc'] = data['median_income']
    data['HouseAge'] = data['housing_median_age']
    data['Latitude'] = data['latitude']
    data['Longitude'] = data['longitude']

    # avg rooms = total rooms / households
    data['AveRooms'] = data['total_rooms'] / data["households"]

    # avg bed rooms = total bed rooms / households
    data['AveBedrms'] = data['total_bedrooms'] / data["households"]

    # avg occupancy = population / households
    data['AveOccup'] = data['population'] / data['households']

    data[targets_column] = data[targets_column] / 100000.0

    if config.cross_val:
        return FoldedDataset(config, data_path=data, features_columns=feature_names, targets_column=targets_column)
    else:
        return NAMDataset(config, data_path=data, features_columns=feature_names, targets_column=targets_column)


def load_gallup_data(config=cfg,
                     data_path: str = 'data/GALLUP.csv',
                     features_columns: list = ["country", "income_2", "WP1219", "WP1220", "year", "weo_gdpc_con_ppp"],
                     targets_column: str = "WP16",
                     weights_column: str = "wgt") -> Dict:

    config.regression = False
    data = pd.read_csv(data_path)
    data["WP16"] = np.where(data["WP16"] < 7, 0, 1)
    # data = data.sample(frac=0.1)

    if config.cross_val:
        return FoldedDataset(config,
                             data_path=data,
                             features_columns=features_columns,
                             targets_column=targets_column,
                             weights_column=weights_column)
    else:
        return NAMDataset(config,
                          data_path=data,
                          features_columns=features_columns,
                          targets_column=targets_column,
                          weights_column=weights_column)



from typing import Tuple
from typing import Union

import pandas as pd
from loguru import logger
from sklearn.model_selection import KFold
from sklearn.model_selection import ShuffleSplit
from sklearn.model_selection import StratifiedKFold
from sklearn.model_selection import StratifiedShuffleSplit
from torch.utils.data import DataLoader
from torch.utils.data import Subset

from .base import NAMDataset


class FoldedDataset(NAMDataset):

    def __init__(self,
                 config,
                 data_path: Union[str, pd.DataFrame],
                 features_columns: list,
                 targets_column: str,
                 weights_column: str = None) -> None:
        super().__init__(config=config,
                         data_path=data_path,
                         features_columns=features_columns,
                         targets_column=targets_column,
                         weights_column=weights_column)

        self.train_subset, self.test_subset = self.get_folds()

    def get_folds(self, shuffle: bool = True, random_state: int = 42) -> Tuple[Subset, ...]:
        fold_num = self.config.fold_num
        n_splits = self.config.num_splits
        stratified = not self.config.regression

        if stratified:
            kf = StratifiedKFold(n_splits=n_splits, shuffle=shuffle, random_state=random_state)
        else:
            kf = KFold(n_splits=n_splits, shuffle=shuffle, random_state=random_state)

        assert fold_num <= n_splits and fold_num > 0, 'Pass a valid fold number.'
        for train_index, test_index in kf.split(self.features, self.targets):
            if fold_num == 1:
                train = Subset(self, train_index)
                test = Subset(self, test_index)
                return train, test
            else:
                fold_num -= 1

    def train_dataloaders(self,
                          test_size: int = 0.125,
                          shuffle: bool = True,
                          random_state: int = 42) -> Tuple[DataLoader, ...]:

        num_folds = self.config.num_folds
        stratified = not self.config.regression
        if stratified:
            shuffle_split = StratifiedShuffleSplit(n_splits=num_folds, test_size=test_size, random_state=random_state)
        else:
            shuffle_split = ShuffleSplit(n_splits=num_folds, test_size=test_size, random_state=random_state)

        for i, (train_index, validation_index) in enumerate(
                shuffle_split.split(self.features[self.train_subset.indices], self.targets[self.train_subset.indices])):

            train = Subset(self.train_subset, train_index)
            val = Subset(self.train_subset, validation_index)

            trainloader = DataLoader(
                train,
                batch_size=self.config.batch_size,
                shuffle=shuffle,
            )
            valloader = DataLoader(
                val,
                batch_size=self.config.batch_size,
                shuffle=False,
            )

            logger.info(f'Fold[{i + 1}]: train: {len(trainloader.dataset)}, val: {len(valloader.dataset)}')

            yield trainloader, valloader

    def test_dataloaders(self) -> DataLoader:
        return DataLoader(self.test_subset, batch_size=self.config.batch_size, shuffle=False)



"""Utilities for logging to Weights & Biases."""

import torch
import wandb

from nam.utils.loggers import base


class WandBLogger(base.Logger):
    """Logs to a `wandb` dashboard."""

    def __init__(
        self,
        project: str = "nam",
        configs: dict = None,
    ) -> None:
        super().__init__(log_dir=configs.logdir)
        wandb.init(project=project, config=configs)

    def write(self, data: base.LoggingData) -> None:
        wandb.log(data)

    def watch(
        self,
        model: torch.nn.Module,
        **kwargs: dict,
    ) -> None:
        wandb.watch(model, kwargs)



import pytorch_lightning as pl
import torch
from torch import nn

from nam.trainer.losses import penalized_loss
from nam.trainer.metrics import accuracy
from nam.trainer.metrics import mae
from nam.types import Config


class LitNAM(pl.LightningModule):

    def __init__(self, config: Config, model: nn.Module) -> None:
        super().__init__()
        self.config = Config(**vars(config))  #config
        self.model = model

        self.criterion = lambda inputs, targets, weights, fnns_out, model: penalized_loss(
            self.config, inputs, targets, weights, fnns_out, model)

        self.metrics = lambda logits, targets: mae(logits, targets) if config.regression else accuracy(logits, targets)
        self.metrics_name = "MAE" if config.regression else "Accuracy"

        self.save_hyperparameters(vars(self.config))

    def forward(self, inputs: torch.Tensor) -> torch.Tensor:
        logits, fnns_out = self.model(inputs)
        return logits, fnns_out

    def configure_optimizers(self):
        optimizer = torch.optim.AdamW(self.parameters(), lr=self.config.lr, weight_decay=self.config.decay_rate)

        scheduler = torch.optim.lr_scheduler.StepLR(optimizer, gamma=0.995, step_size=1)
        return [optimizer], [scheduler]

    def training_step(self, batch, batch_idx):
        features, targets = batch
        # inputs, targets, *weights = batch
        # weights = weights.pop() if weights else torch.tensor(1)

        logits, fnns_out = self.model(features)
        loss = self.criterion(logits, targets, None, fnns_out, self.model)
        metric = self.metrics(logits, targets)

        self.log_dict(
            {
                'train_loss': loss,
                f"{self.metrics_name}_metric": metric
            },
            on_step=True,
            on_epoch=True,
            prog_bar=True,
            logger=True,
        )

        return loss  #{'training_loss': loss}

    def validation_step(self, batch, batch_idx):
        features, targets = batch
        # inputs, targets, *weights = batch
        # weights = weights.pop() if weights else torch.tensor(1)

        logits, fnns_out = self.model(features)  #, weights)
        loss = self.criterion(logits, targets, None, fnns_out, self.model)
        metric = self.metrics(logits, targets)

        self.log_dict(
            {
                'val_loss': loss,
                f"{self.metrics_name}_metric": metric
            },
            on_step=True,
            on_epoch=True,
            prog_bar=True,
            logger=True,
        )

        return {'val_loss': loss}

    def test_step(self, batch, batch_idx):
        features, targets = batch
        # inputs, targets, *weights = batch

        logits, fnns_out = self.model(features)
        loss = self.criterion(logits, targets, None, fnns_out, self.model)
        metric = self.metrics(logits, targets)

        self.log_dict(
            {
                'test_loss': loss,
                f"{self.metrics_name}_metric": metric
            },
            on_step=True,
            on_epoch=True,
            prog_bar=True,
            logger=True,
        )

        return {'test_loss': loss}



from typing import List

import numpy as np
import torch
import torch.nn as nn


def init_weights(m):
    if type(m) == nn.Linear:
        torch.nn.init.kaiming_normal_(m.weight)
        m.bias.data.fill_(0.01)


def get_num_units(
    config,
    features: torch.Tensor,
) -> List:
    features = features.cpu()
    num_unique_vals = [len(np.unique(features[:, i])) for i in range(features.shape[1])]

    num_units = [min(config.num_basis_functions, i * config.units_multiplier) for i in num_unique_vals]

    return num_units



from setuptools import find_packages
from setuptools import setup

with open("README.md", "r") as f:
    long_description = f.read()

setup(
    name='nam',
    version='0.0.3',
    description="Neural Additive Models (Google Research)",
    long_description=long_description,
    long_description_content_type="text/markdown",
    author="Amr Kayid",
    url="https://github.com/AmrMKayid/nam",
    packages=find_packages(),
    install_requires=[
        "torch",
        "pytorch-lightning",
        "wandb",
        "matplotlib",
        "sklearn",
        "numpy",
        "pandas",
        "loguru",
    ],
    classifiers=[
        "Programming Language :: Python :: 3.6",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
)



"""Utilities for logging to the terminal."""

import os
import time

import wandb
from torch.utils.tensorboard import SummaryWriter

from nam.utils.loggers import base


def _format_key(key: str) -> str:
    """Internal function for formatting keys in Tensorboard format."""
    return key.title().replace('_', '')


class TensorBoardLogger(base.Logger):
    """A simple `Pytorch`-friendly `TensorBoard` wrapper."""

    def __init__(self, config: dict = None, project: str = "nam", label: str = 'Logs') -> None:
        """Initializes the logger. Constructs a simple `TensorBoard` wrapper.

        Args:
          logdir: directory to which we should log files.
          label: label string to use when logging. Default to 'Logs'.
        """
        self._time = time.time()
        self.label = label
        self.config = config
        self._iter = 0

        # Makes sure output directories exist.
        log_dir = os.path.join(config.logdir, "logs")
        os.makedirs(log_dir, exist_ok=True)

        # Initialise the `TensorBoard` writers.
        self._summary_writter = SummaryWriter(log_dir=log_dir)

        if config.wandb:
            wandb.init(project=project, config=vars(config), reinit=True, magic=True)

    def write(self, values: base.LoggingData):
        for key, value in values.items():
            self._summary_writter.add_scalar(f'{self.label}/{_format_key(key)}', value, global_step=self._iter)
            if self.config.wandb:
                wandb.log({f'{self.label}/{_format_key(key)}': value})

        self._iter += 1



from types import SimpleNamespace


class Config(SimpleNamespace):
    """Wrapper around SimpleNamespace, allows dot notation attribute access."""

    @staticmethod
    def map_entry(entry):
        if isinstance(entry, dict):
            return Config(**entry)

        return entry

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        for key, val in kwargs.items():
            if type(val) == dict:
                setattr(self, key, Config(**val))
            elif type(val) == list:
                setattr(self, key, list(map(self.map_entry, val)))

    def update(self, **kwargs):
        for key, val in kwargs.items():
            if type(val) == dict:
                setattr(self, key, Config(**val))
            elif type(val) == list:
                setattr(self, key, list(map(self.map_entry, val)))
            else:
                setattr(self, key, val)



import torch
import torch.nn.functional as F
from torch.nn.parameter import Parameter


class ExU(torch.nn.Module):

    def __init__(
        self,
        in_features: int,
        out_features: int,
    ) -> None:
        super(ExU, self).__init__()
        self.in_features = in_features
        self.out_features = out_features
        self.weights = Parameter(torch.Tensor(in_features, out_features))
        self.bias = Parameter(torch.Tensor(in_features))
        self.reset_parameters()

    def reset_parameters(self) -> None:
        ## Page(4): initializing the weights using a normal distribution
        ##          N(x; 0:5) with x 2 [3; 4] works well in practice.
        torch.nn.init.trunc_normal_(self.weights, mean=4.0, std=0.5)
        torch.nn.init.trunc_normal_(self.bias, std=0.5)

    def forward(
        self,
        inputs: torch.Tensor,
        n: int = 1,
    ) -> torch.Tensor:
        output = (inputs - self.bias).matmul(torch.exp(self.weights))

        # ReLU activations capped at n (ReLU-n)
        output = F.relu(output)
        output = torch.clamp(output, 0, n)

        return output

    def extra_repr(self):
        return f'in_features={self.in_features}, out_features={self.out_features}'



import torch
import torch.nn as nn

from nam.models.base import Model
from nam.models.utils import init_weights


class DNN(Model):

    def __init__(
        self,
        config,
        name: str = "DNNModel",
        *,
        input_shape: int = 1,
        output_shape: int = 1,
    ) -> None:
        super(DNN, self).__init__(config, name)

        self.layers = []
        self.dropout = nn.Dropout(p=self.config.dropout)
        hidden_sizes = self.config.hidden_sizes

        self.layers.append(nn.Linear(input_shape, hidden_sizes[0], bias=True))
        self.layers.append(nn.ReLU())
        self.layers.append(self.dropout)

        for in_features, out_features in zip(hidden_sizes, hidden_sizes[1:]):
            self.layers.append(nn.Linear(in_features, out_features, bias=True))
            self.layers.append(nn.ReLU())
            self.layers.append(self.dropout)

        self.layers.append(nn.Linear(hidden_sizes[-1], output_shape, bias=True))

        self.model = nn.Sequential(*self.layers)
        self.apply(init_weights)

    def forward(self, inputs) -> torch.Tensor:
        return self.model(inputs), None



import torch
import torch.nn as nn
import torch.nn.functional as F
from torch.nn.parameter import Parameter


class LinReLU(torch.nn.Module):
    __constants__ = ['bias']

    def __init__(
        self,
        in_features: int,
        out_features: int,
    ) -> None:
        super(LinReLU, self).__init__()
        self.in_features = in_features
        self.out_features = out_features
        self.weights = Parameter(torch.Tensor(in_features, out_features))
        self.bias = Parameter(torch.Tensor(in_features))

        self.reset_parameters()

    def reset_parameters(self) -> None:
        nn.init.xavier_uniform_(self.weights)
        torch.nn.init.trunc_normal_(self.bias, std=0.5)

    def forward(
        self,
        inputs: torch.Tensor,
    ) -> torch.Tensor:
        output = (inputs - self.bias) @ self.weights
        output = F.relu(output)

        return output

    def extra_repr(self):
        return f'in_features={self.in_features}, out_features={self.out_features}'
